@@ -12,54 +12,85 @@
  *
  * Licensed under GPLv2 or later, see file LICENSE in this source tree.
  */
+//config:config STAT
+//config:	bool "stat (11 kb)"
+//config:	default y
+//config:	help
+//config:	display file or filesystem status.
+//config:
+//config:config FEATURE_STAT_FORMAT
+//config:	bool "Enable custom formats (-c)"
+//config:	default y
+//config:	depends on STAT
+//config:	help
+//config:	Without this, stat will not support the '-c format' option where
+//config:	users can pass a custom format string for output. This adds about
+//config:	7k to a nonstatic build on amd64.
+//config:
+//config:config FEATURE_STAT_FILESYSTEM
+//config:	bool "Enable display of filesystem status (-f)"
+//config:	default y
+//config:	depends on STAT
+//config:	help
+//config:	Without this, stat will not support the '-f' option to display
+//config:	information about filesystem status.
+
+//applet:IF_STAT(APPLET_NOEXEC(stat, stat, BB_DIR_BIN, BB_SUID_DROP, stat))
+
+//kbuild:lib-$(CONFIG_STAT) += stat.o
 
 //usage:#define stat_trivial_usage
-//usage:       "[OPTIONS] FILE..."
+//usage:       "[-lt"IF_FEATURE_STAT_FILESYSTEM("f")"] "IF_FEATURE_STAT_FORMAT("[-c FMT] ")"FILE..."
 //usage:#define stat_full_usage "\n\n"
-//usage:       "Display file (default) or filesystem status\n"
+//usage:       "Display file"
+//usage:            IF_FEATURE_STAT_FILESYSTEM(" (default) or filesystem")
+//usage:            " status\n"
 //usage:	IF_FEATURE_STAT_FORMAT(
-//usage:     "\n	-c fmt	Use the specified format"
+//usage:     "\n	-c FMT	Use the specified format"
 //usage:	)
+//usage:	IF_FEATURE_STAT_FILESYSTEM(
 //usage:     "\n	-f	Display filesystem status"
+//usage:	)
 //usage:     "\n	-L	Follow links"
-//usage:     "\n	-t	Display info in terse form"
+//usage:     "\n	-t	Terse display"
 //usage:	IF_SELINUX(
 //usage:     "\n	-Z	Print security context"
 //usage:	)
 //usage:	IF_FEATURE_STAT_FORMAT(
-//usage:       "\n\nValid format sequences for files:\n"
+//usage:       "\n\nFMT sequences"IF_FEATURE_STAT_FILESYSTEM(" for files")":\n"
 //usage:       " %a	Access rights in octal\n"
 //usage:       " %A	Access rights in human readable form\n"
 //usage:       " %b	Number of blocks allocated (see %B)\n"
-//usage:       " %B	The size in bytes of each block reported by %b\n"
+//usage:       " %B	Size in bytes of each block reported by %b\n"
 //usage:       " %d	Device number in decimal\n"
 //usage:       " %D	Device number in hex\n"
 //usage:       " %f	Raw mode in hex\n"
 //usage:       " %F	File type\n"
-//usage:       " %g	Group ID of owner\n"
-//usage:       " %G	Group name of owner\n"
+//usage:       " %g	Group ID\n"
+//usage:       " %G	Group name\n"
 //usage:       " %h	Number of hard links\n"
 //usage:       " %i	Inode number\n"
 //usage:       " %n	File name\n"
 //usage:       " %N	File name, with -> TARGET if symlink\n"
 //usage:       " %o	I/O block size\n"
-//usage:       " %s	Total size, in bytes\n"
+//usage:       " %s	Total size in bytes\n"
 //usage:       " %t	Major device type in hex\n"
 //usage:       " %T	Minor device type in hex\n"
-//usage:       " %u	User ID of owner\n"
-//usage:       " %U	User name of owner\n"
+//usage:       " %u	User ID\n"
+//usage:       " %U	User name\n"
 //usage:       " %x	Time of last access\n"
 //usage:       " %X	Time of last access as seconds since Epoch\n"
 //usage:       " %y	Time of last modification\n"
 //usage:       " %Y	Time of last modification as seconds since Epoch\n"
 //usage:       " %z	Time of last change\n"
 //usage:       " %Z	Time of last change as seconds since Epoch\n"
-//usage:       "\nValid format sequences for file systems:\n"
+//usage:	IF_FEATURE_STAT_FILESYSTEM(
+//usage:       "\nFMT sequences for file systems:\n"
 //usage:       " %a	Free blocks available to non-superuser\n"
-//usage:       " %b	Total data blocks in file system\n"
-//usage:       " %c	Total file nodes in file system\n"
-//usage:       " %d	Free file nodes in file system\n"
-//usage:       " %f	Free blocks in file system\n"
+//usage:       " %b	Total data blocks\n"
+//usage:       " %c	Total file nodes\n"
+//usage:       " %d	Free file nodes\n"
+//usage:       " %f	Free blocks\n"
 //usage:	IF_SELINUX(
 //usage:       " %C	Security context in selinux\n"
 //usage:	)
@@ -71,13 +102,17 @@
 //usage:       " %t	Type in hex\n"
 //usage:       " %T	Type in human readable form"
 //usage:	)
+//usage:	)
 
 #include "libbb.h"
+#include "common_bufsiz.h"
 
-#define OPT_FILESYS     (1 << 0)
-#define OPT_TERSE       (1 << 1)
-#define OPT_DEREFERENCE (1 << 2)
-#define OPT_SELINUX     (1 << 3)
+enum {
+	OPT_TERSE       = (1 << 0),
+	OPT_DEREFERENCE = (1 << 1),
+	OPT_FILESYS     = (1 << 2) * ENABLE_FEATURE_STAT_FILESYSTEM,
+	OPT_SELINUX     = (1 << (2+ENABLE_FEATURE_STAT_FILESYSTEM)) * ENABLE_SELINUX,
+};
 
 #if ENABLE_FEATURE_STAT_FORMAT
 typedef bool (*statfunc_ptr)(const char *, const char *);
@@ -114,24 +149,57 @@ static const char *file_type(const struct stat *st)
 	return "weird file";
 }
 
-static const char *human_time(time_t t)
+static const char *human_time(struct timespec *ts)
 {
-	/* Old
-	static char *str;
-	str = ctime(&t);
-	str[strlen(str)-1] = '\0';
-	return str;
-	*/
-	/* coreutils 6.3 compat: */
+	char fmt[sizeof("%Y-%m-%d %H:%M:%S.123456789 %z") + /*paranoia*/ 8];
 
-	/*static char buf[sizeof("YYYY-MM-DD HH:MM:SS.000000000")] ALIGN1;*/
+	/* coreutils 6.3 compat */
 #define buf bb_common_bufsiz1
+	setup_common_bufsiz();
 
-	strcpy(strftime_YYYYMMDDHHMMSS(buf, sizeof(buf), &t), ".000000000");
+	sprintf(stpcpy(fmt, "%Y-%m-%d %H:%M:%S"), ".%09u %%z", (unsigned)ts->tv_nsec);
+	strftime(buf, COMMON_BUFSIZE, fmt, localtime(&ts->tv_sec));
 	return buf;
 #undef buf
 }
 
+#if ENABLE_FEATURE_STAT_FILESYSTEM
+#define FS_TYPE_LIST \
+FS_TYPE(0xADFF,     "affs") \
+FS_TYPE(0x1CD1,     "devpts") \
+FS_TYPE(0x137D,     "ext") \
+FS_TYPE(0xEF51,     "ext2") \
+FS_TYPE(0xEF53,     "ext2/ext3") \
+FS_TYPE(0x3153464a, "jfs") \
+FS_TYPE(0x58465342, "xfs") \
+FS_TYPE(0xF995E849, "hpfs") \
+FS_TYPE(0x9660,     "isofs") \
+FS_TYPE(0x4000,     "isofs") \
+FS_TYPE(0x4004,     "isofs") \
+FS_TYPE(0x137F,     "minix") \
+FS_TYPE(0x138F,     "minix (30 char.)") \
+FS_TYPE(0x2468,     "minix v2") \
+FS_TYPE(0x2478,     "minix v2 (30 char.)") \
+FS_TYPE(0x4d44,     "msdos") \
+FS_TYPE(0x4006,     "fat") \
+FS_TYPE(0x564c,     "novell") \
+FS_TYPE(0x6969,     "nfs") \
+FS_TYPE(0x9fa0,     "proc") \
+FS_TYPE(0x517B,     "smb") \
+FS_TYPE(0x012FF7B4, "xenix") \
+FS_TYPE(0x012FF7B5, "sysv4") \
+FS_TYPE(0x012FF7B6, "sysv2") \
+FS_TYPE(0x012FF7B7, "coh") \
+FS_TYPE(0x00011954, "ufs") \
+FS_TYPE(0x012FD16D, "xia") \
+FS_TYPE(0x5346544e, "ntfs") \
+FS_TYPE(0x1021994,  "tmpfs") \
+FS_TYPE(0x52654973, "reiserfs") \
+FS_TYPE(0x28cd3d45, "cramfs") \
+FS_TYPE(0x7275,     "romfs") \
+FS_TYPE(0x858458f6, "ramfs") \
+FS_TYPE(0x73717368, "squashfs") \
+FS_TYPE(0x62656572, "sysfs")
 /* Return the type of the specified file system.
  * Some systems have statfvs.f_basetype[FSTYPSZ]. (AIX, HP-UX, and Solaris)
  * Others have statfs.f_fstypename[MFSNAMELEN]. (NetBSD 1.5.2)
@@ -139,54 +207,22 @@ static const char *human_time(time_t t)
  */
 static const char *human_fstype(uint32_t f_type)
 {
-	static const struct types {
-		uint32_t type;
-		const char *const fs;
-	} humantypes[] = {
-		{ 0xADFF,     "affs" },
-		{ 0x1Cd1,     "devpts" },
-		{ 0x137D,     "ext" },
-		{ 0xEF51,     "ext2" },
-		{ 0xEF53,     "ext2/ext3" },
-		{ 0x3153464a, "jfs" },
-		{ 0x58465342, "xfs" },
-		{ 0xF995E849, "hpfs" },
-		{ 0x9660,     "isofs" },
-		{ 0x4000,     "isofs" },
-		{ 0x4004,     "isofs" },
-		{ 0x137F,     "minix" },
-		{ 0x138F,     "minix (30 char.)" },
-		{ 0x2468,     "minix v2" },
-		{ 0x2478,     "minix v2 (30 char.)" },
-		{ 0x4d44,     "msdos" },
-		{ 0x4006,     "fat" },
-		{ 0x564c,     "novell" },
-		{ 0x6969,     "nfs" },
-		{ 0x9fa0,     "proc" },
-		{ 0x517B,     "smb" },
-		{ 0x012FF7B4, "xenix" },
-		{ 0x012FF7B5, "sysv4" },
-		{ 0x012FF7B6, "sysv2" },
-		{ 0x012FF7B7, "coh" },
-		{ 0x00011954, "ufs" },
-		{ 0x012FD16D, "xia" },
-		{ 0x5346544e, "ntfs" },
-		{ 0x1021994,  "tmpfs" },
-		{ 0x52654973, "reiserfs" },
-		{ 0x28cd3d45, "cramfs" },
-		{ 0x7275,     "romfs" },
-		{ 0x858458f6, "romfs" },
-		{ 0x73717368, "squashfs" },
-		{ 0x62656572, "sysfs" },
-		{ 0, "UNKNOWN" }
+# define FS_TYPE(type, name) type,
+	static const uint32_t fstype[] ALIGN4 = {
+		FS_TYPE_LIST
 	};
-
+# undef FS_TYPE
+# define FS_TYPE(type, name) name"\0"
+	static const char humanname[] ALIGN1 =
+		FS_TYPE_LIST
+		"UNKNOWN";
+# undef FS_TYPE
 	int i;
 
-	for (i = 0; humantypes[i].type; ++i)
-		if (humantypes[i].type == f_type)
+	for (i = 0; i < ARRAY_SIZE(fstype); ++i)
+		if (fstype[i] == f_type)
 			break;
-	return humantypes[i].fs;
+	return nth_string(humanname, i);
 }
 
 /* "man statfs" says that statfsbuf->f_fsid is a mess */
@@ -202,6 +238,7 @@ static unsigned long long get_f_fsid(const struct statfs *statfsbuf)
 	while (--sz > 0);
 	return r;
 }
+#endif  /* FEATURE_STAT_FILESYSTEM */
 
 #if ENABLE_FEATURE_STAT_FORMAT
 static void strcatc(char *str, char c)
@@ -217,6 +254,7 @@ static void printfs(char *pformat, const char *msg)
 	printf(pformat, msg);
 }
 
+#if ENABLE_FEATURE_STAT_FILESYSTEM
 /* print statfs info */
 static void FAST_FUNC print_statfs(char *pformat, const char m,
 		const char *const filename, const void *data
@@ -263,6 +301,7 @@ static void FAST_FUNC print_statfs(char *pformat, const char m,
 		printf(pformat, m);
 	}
 }
+#endif
 
 /* print stat info */
 static void FAST_FUNC print_stat(char *pformat, const char m,
@@ -300,7 +339,8 @@ static void FAST_FUNC print_stat(char *pformat, const char m,
 		strcat(pformat, "lo");
 		printf(pformat, (unsigned long) (statbuf->st_mode & (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)));
 	} else if (m == 'A') {
-		printfs(pformat, bb_mode_string(statbuf->st_mode));
+		char modestr[12];
+		printfs(pformat, bb_mode_string(modestr, statbuf->st_mode));
 	} else if (m == 'f') {
 		strcat(pformat, "lx");
 		printf(pformat, (unsigned long) statbuf->st_mode);
@@ -340,19 +380,19 @@ static void FAST_FUNC print_stat(char *pformat, const char m,
 		strcat(pformat, "lu");
 		printf(pformat, (unsigned long) statbuf->st_blksize);
 	} else if (m == 'x') {
-		printfs(pformat, human_time(statbuf->st_atime));
+		printfs(pformat, human_time(&statbuf->st_atim));
 	} else if (m == 'X') {
 		strcat(pformat, TYPE_SIGNED(time_t) ? "ld" : "lu");
 		/* note: (unsigned long) would be wrong:
 		 * imagine (unsigned long64)int32 */
 		printf(pformat, (long) statbuf->st_atime);
 	} else if (m == 'y') {
-		printfs(pformat, human_time(statbuf->st_mtime));
+		printfs(pformat, human_time(&statbuf->st_mtim));
 	} else if (m == 'Y') {
 		strcat(pformat, TYPE_SIGNED(time_t) ? "ld" : "lu");
 		printf(pformat, (long) statbuf->st_mtime);
 	} else if (m == 'z') {
-		printfs(pformat, human_time(statbuf->st_ctime));
+		printfs(pformat, human_time(&statbuf->st_ctim));
 	} else if (m == 'Z') {
 		strcat(pformat, TYPE_SIGNED(time_t) ? "ld" : "lu");
 		printf(pformat, (long) statbuf->st_ctime);
@@ -374,7 +414,7 @@ static void print_it(const char *masterformat,
 {
 	/* Create a working copy of the format string */
 	char *format = xstrdup(masterformat);
-	/* Add 2 to accomodate our conversion of the stat '%s' format string
+	/* Add 2 to accommodate our conversion of the stat '%s' format string
 	 * to the printf '%llu' one.  */
 	char *dest = xmalloc(strlen(format) + 2 + 1);
 	char *b;
@@ -400,7 +440,7 @@ static void print_it(const char *masterformat,
 
 		/* print preceding string */
 		*p = '\0';
-		fputs(b, stdout);
+		fputs_stdout(b);
 
 		p += len;
 		b = p + 1;
@@ -423,6 +463,7 @@ static void print_it(const char *masterformat,
 }
 #endif  /* FEATURE_STAT_FORMAT */
 
+#if ENABLE_FEATURE_STAT_FILESYSTEM
 /* Stat the file system and print what we find.  */
 #if !ENABLE_FEATURE_STAT_FORMAT
 #define do_statfs(filename, format) do_statfs(filename)
@@ -456,7 +497,7 @@ static bool do_statfs(const char *filename, const char *format)
 	if (format == NULL) {
 # if !ENABLE_SELINUX
 		format = (option_mask32 & OPT_TERSE
-			? "%n %i %l %t %s %b %f %a %c %d\n"
+			? "%n %i %l %t %s %b %f %a %c %d"
 			: "  File: \"%n\"\n"
 			  "    ID: %-8i Namelen: %-7l Type: %T\n"
 			  "Block size: %-10s\n"
@@ -464,25 +505,26 @@ static bool do_statfs(const char *filename, const char *format)
 			  "Inodes: Total: %-10c Free: %d");
 # else
 		format = (option_mask32 & OPT_TERSE
-			? (option_mask32 & OPT_SELINUX ? "%n %i %l %t %s %b %f %a %c %d %C\n":
-			"%n %i %l %t %s %b %f %a %c %d\n")
-			: (option_mask32 & OPT_SELINUX ?
-			"  File: \"%n\"\n"
-			"    ID: %-8i Namelen: %-7l Type: %T\n"
-			"Block size: %-10s\n"
-			"Blocks: Total: %-10b Free: %-10f Available: %a\n"
-			"Inodes: Total: %-10c Free: %d"
-			"  S_context: %C\n":
-			"  File: \"%n\"\n"
-			"    ID: %-8i Namelen: %-7l Type: %T\n"
-			"Block size: %-10s\n"
-			"Blocks: Total: %-10b Free: %-10f Available: %a\n"
-			"Inodes: Total: %-10c Free: %d\n")
+			? (option_mask32 & OPT_SELINUX
+				? "%n %i %l %t %s %b %f %a %c %d %C"
+				: "%n %i %l %t %s %b %f %a %c %d")
+			: (option_mask32 & OPT_SELINUX
+				? "  File: \"%n\"\n"
+				"    ID: %-8i Namelen: %-7l Type: %T\n"
+				"Block size: %-10s\n"
+				"Blocks: Total: %-10b Free: %-10f Available: %a\n"
+				"Inodes: Total: %-10c Free: %d"
+				"  S_context: %C"
+				: "  File: \"%n\"\n"
+				"    ID: %-8i Namelen: %-7l Type: %T\n"
+				"Block size: %-10s\n"
+				"Blocks: Total: %-10b Free: %-10f Available: %a\n"
+				"Inodes: Total: %-10c Free: %d")
 			);
 # endif /* SELINUX */
 	}
 	print_it(format, filename, print_statfs, &statfsbuf IF_SELINUX(, scontext));
-#else /* FEATURE_STAT_FORMAT */
+#else /* !FEATURE_STAT_FORMAT */
 	format = (option_mask32 & OPT_TERSE
 		? "%s %llx %lu "
 		: "  File: \"%s\"\n"
@@ -538,6 +580,7 @@ static bool do_statfs(const char *filename, const char *format)
 #endif  /* FEATURE_STAT_FORMAT */
 	return 1;
 }
+#endif  /* FEATURE_STAT_FILESYSTEM */
 
 /* stat the file and print what we find */
 #if !ENABLE_FEATURE_STAT_FORMAT
@@ -578,14 +621,14 @@ static bool do_stat(const char *filename, const char *format)
 					"Device: %Dh/%dd\tInode: %-10i  Links: %-5h"
 					" Device type: %t,%T\n"
 					"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
-					"Access: %x\n" "Modify: %y\n" "Change: %z\n";
+					"Access: %x\n" "Modify: %y\n" "Change: %z";
 			} else {
 				format =
 					"  File: %N\n"
 					"  Size: %-10s\tBlocks: %-10b IO Block: %-6o %F\n"
 					"Device: %Dh/%dd\tInode: %-10i  Links: %h\n"
 					"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
-					"Access: %x\n" "Modify: %y\n" "Change: %z\n";
+					"Access: %x\n" "Modify: %y\n" "Change: %z";
 			}
 		}
 # else
@@ -604,14 +647,14 @@ static bool do_stat(const char *filename, const char *format)
 					" Device type: %t,%T\n"
 					"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
 					"   S_Context: %C\n"
-					"Access: %x\n" "Modify: %y\n" "Change: %z\n"
+					"Access: %x\n" "Modify: %y\n" "Change: %z"
 					:
 					"  File: %N\n"
 					"  Size: %-10s\tBlocks: %-10b IO Block: %-6o %F\n"
 					"Device: %Dh/%dd\tInode: %-10i  Links: %-5h"
 					" Device type: %t,%T\n"
 					"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
-					"Access: %x\n" "Modify: %y\n" "Change: %z\n"
+					"Access: %x\n" "Modify: %y\n" "Change: %z"
 					);
 			} else {
 				format = (option_mask32 & OPT_SELINUX ?
@@ -620,13 +663,13 @@ static bool do_stat(const char *filename, const char *format)
 					"Device: %Dh/%dd\tInode: %-10i  Links: %h\n"
 					"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
 					"S_Context: %C\n"
-					"Access: %x\n" "Modify: %y\n" "Change: %z\n"
+					"Access: %x\n" "Modify: %y\n" "Change: %z"
 					:
 					"  File: %N\n"
 					"  Size: %-10s\tBlocks: %-10b IO Block: %-6o %F\n"
 					"Device: %Dh/%dd\tInode: %-10i  Links: %h\n"
 					"Access: (%04a/%10.10A)  Uid: (%5u/%8U)   Gid: (%5g/%8G)\n"
-					"Access: %x\n" "Modify: %y\n" "Change: %z\n"
+					"Access: %x\n" "Modify: %y\n" "Change: %z"
 					);
 			}
 		}
@@ -660,6 +703,7 @@ static bool do_stat(const char *filename, const char *format)
 			bb_putchar('\n');
 # endif
 	} else {
+		char modestr[12];
 		char *linkname = NULL;
 		struct passwd *pw_ent;
 		struct group *gw_ent;
@@ -694,7 +738,7 @@ static bool do_stat(const char *filename, const char *format)
 			bb_putchar('\n');
 		printf("Access: (%04lo/%10.10s)  Uid: (%5lu/%8s)   Gid: (%5lu/%8s)\n",
 		       (unsigned long) (statbuf.st_mode & (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)),
-		       bb_mode_string(statbuf.st_mode),
+		       bb_mode_string(modestr, statbuf.st_mode),
 		       (unsigned long) statbuf.st_uid,
 		       (pw_ent != NULL) ? pw_ent->pw_name : "UNKNOWN",
 		       (unsigned long) statbuf.st_gid,
@@ -703,9 +747,9 @@ static bool do_stat(const char *filename, const char *format)
 		if (option_mask32 & OPT_SELINUX)
 			printf("   S_Context: %s\n", scontext);
 # endif
-		printf("Access: %s\n", human_time(statbuf.st_atime));
-		printf("Modify: %s\n", human_time(statbuf.st_mtime));
-		printf("Change: %s\n", human_time(statbuf.st_ctime));
+		printf("Access: %s\n", human_time(&statbuf.st_atim));
+		printf("Modify: %s\n", human_time(&statbuf.st_mtim));
+		printf("Change: %s\n", human_time(&statbuf.st_ctim));
 	}
 #endif  /* FEATURE_STAT_FORMAT */
 	return 1;
@@ -717,16 +761,24 @@ int stat_main(int argc UNUSED_PARAM, char **argv)
 	IF_FEATURE_STAT_FORMAT(char *format = NULL;)
 	int i;
 	int ok;
-	unsigned opts;
 	statfunc_ptr statfunc = do_stat;
+#if ENABLE_FEATURE_STAT_FILESYSTEM || ENABLE_SELINUX
+	unsigned opts;
 
-	opt_complementary = "-1"; /* min one arg */
-	opts = getopt32(argv, "ftL"
+	opts =
+#endif
+	getopt32(argv, "^"
+		"tL"
+		IF_FEATURE_STAT_FILESYSTEM("f")
 		IF_SELINUX("Z")
-		IF_FEATURE_STAT_FORMAT("c:", &format)
+		IF_FEATURE_STAT_FORMAT("c:")
+		"\0" "-1" /* min one arg */
+		IF_FEATURE_STAT_FORMAT(,&format)
 	);
+#if ENABLE_FEATURE_STAT_FILESYSTEM
 	if (opts & OPT_FILESYS) /* -f */
 		statfunc = do_statfs;
+#endif
 #if ENABLE_SELINUX
 	if (opts & OPT_SELINUX) {
 		selinux_or_die();
